;=====================================================
;VARIOUS PROCEDURES
;=====================================================
;Note: for some of the 3D procedures, the fcos instruction is
;uselessly followed by a fabs (absolute value) instruction, which
;is unnecessary due to nature of cosine.
;Note2: Generally, "STD" is short for "Standard".

;============================================================
CheckCharsForNull4:	;eax = a DWORD of 4 chars
;============================================================
;destroys: eax,ecx
;returns in eax the same value but with all null chars converted to spaces

	mov ecx,4
	.scanloop:
	test al,al
	     jnz @f
	     mov al,20h
	@@:
	rol eax,8
	dec ecx
	    jnz .scanloop
ret

;============================================================
OpenReadFullFile:   ;param1=filename, param2=minimum file size, param3 = max file size (-1 if none/4GB)
;============================================================
;returns pointer to buffer in eax if success, null if failure
;returns buffer size in ecx if success

push ebp
mov ebp,esp
sub esp,12
push ebx

	xor ebx,ebx	     ;file handle
	mov [ebp-4],ebx      ;memory pointer
	mov [ebp-8],ebx      ;size of file/memory space
	mov [ebp-12],ebx     ;number of bytes read
	invoke CreateFile,[ebp+8],GENERIC_READ,FILE_SHARE_READ,0,\
			  OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
	cmp eax,-1
	    jz .failed
	mov ebx,eax

	invoke GetFileSize,ebx,0
	cmp eax,-1
	    jz .failed
	cmp eax,[ebp+16]
	    ja .failed
	cmp eax,[ebp+12]
	    jb .failed

	mov [ebp-8],eax

	invoke SetFilePointer,ebx,0,0,FILE_BEGIN
	cmp eax,-1
	    jz .failed

	invoke LocalAlloc,LMEM_FIXED,[ebp-8]
	test eax,eax
	mov [ebp-4],eax
	     jz .failed

	lea edx,[ebp-12]
	invoke ReadFile,ebx,eax,[ebp-8],edx,0
	test eax,eax
	mov ecx,[ebp-12]
	     jz .failed
	cmp ecx,[ebp-8]
	    jnz .failed

	invoke CloseHandle,ebx

	mov eax,[ebp-4]
	mov ecx,[ebp-8]

pop ebx
mov esp,ebp
pop ebp
ret 12

.failed:
	CLOSEHANDLE ebx
	LOCALRELEASENONULL DWORD [ebp-4]
	xor eax,eax
	xor ecx,ecx
pop ebx
mov esp,ebp
pop ebp
ret 12

align 16
;============================================================
GetPlayerPrivAIType:	  ;in eax
;============================================================

	mov eax,[PLAYER_OBJ_INVUL_ADDR]
	mov edx,[eax]		;edi = main "Player" class
	mov eax,[edx+0ACh]	;eax=first sub-class leading to AIType
	mov ecx,[eax+28h]	;ecx=2nd...
	xor eax,eax
	mov edx,[ecx+10h]	;edx= AIType sub-class
	mov al,[edx+08]

ret

align 16
;============================================================
FindNextSaveGameFileName:      ;[esp+4] = buffer in which to return (should be min. 16 bytes)
;============================================================
;if error, first character in buffer will be null
;destroyed: eax,ecx,edx
push ebx
push edi
push esi

	mov ecx,[PathStringSize]
	mov esi,StringBuffer
	mov edi,FindNextSaveGame_BUF
	rep movsb
	mov DWORD [edi],'SAVE'
	mov DWORD [edi+4],'GAME'
	mov BYTE [edi+8],'.'
	add edi,9

	mov esi,[esp+4+12]
	xor ebx,ebx
	mov BYTE [esi],0

	.mainloop:
	invoke wsprintf,edi,FindNextSaveGame_COUNT_STR,ebx
	add esp,12

	invoke FindFirstFile,FindNextSaveGame_BUF,QUICKSAVE_finddatabuf
	cmp eax,INVALID_HANDLE_VALUE
	    jz .namefound

	invoke FindClose,eax

	inc ebx
	cmp ebx,1000
	    jb .mainloop

	xor eax,eax
pop esi
pop edi
pop ebx
ret 4

.namefound:
	 mov edx,[edi]			;savegame extension + null-terminator
	 mov DWORD [esi],'SAVE'
	 mov DWORD [esi+4],'GAME'
	 mov BYTE [esi+8],'.'
	 mov [esi+9],edx

	 mov eax,TRUE
pop esi
pop edi
pop ebx
ret 4

align 16
;============================================================
VirtualProtectNOEX:
;============================================================
	mov eax,[esp]		;return address
	add esp,4		;remove return address
	mov [esp],eax		;replace first param (process handle) with address
	jmp [VirtualProtect]

align 16
;============================================================
GetTickCountINGAME:	   ;returns tick count from level start
;============================================================
;returns tick count, but substracts total time spent in in-game menu
;from tick

	invoke GetTickCountTRUE
	sub eax,[NotInGame_TOTALMENUTIME]
	sub eax,[LevelStart_TICK]
ret

align 16
;============================================================
StartInGameMenuTimer:
;============================================================
;only starts timer if not already not in-game (if in-game)
cmp [IsInGame_BOOL],0		  ;only start not-in-game tick if previously in-game
    jz .exitproc

push eax
push ecx
push edx
	invoke GetTickCountTRUE
	mov [NotInGameStart_TICK],eax
	mov [NotInGame_BOOL],1
pop edx
pop ecx
pop eax
.exitproc:
ret

align 16
;============================================================
IsPlayerHoldingCGun:  ;destroyed: eax,edx, returns 0 if false or CGun object otherwise
;============================================================
	mov eax,[ActiveGunRef]
	mov edx,[eax]
	mov eax,[edx]
	mov edx,[eax+08h]
	mov eax,[edx+GETWEAPON_PRIMARY]     ;eax = possible primary CGun object (object held by Anne)
	test eax,eax
	mov edx,[CGUNVTABLE]
	     jz .exitproc		    ;make sure player is holding an object
	cmp [eax+CInstance.lpVTable],edx
	    jz .exitproc
	xor eax,eax
.exitproc:
ret

align 16
;============================================================
GetCPlayerPrivInstance: 	;destroyed: ecx only, returns in ecx
;============================================================
       mov ecx,[PLAYER_OBJ_INVUL_ADDR]
       mov ecx,[ecx]
ret

align 16
;============================================================
GetInstanceFromString:	 ;eax = instance's name, returns in eax (null if error)
;============================================================
;THIS FUNCTION IS COPIED ALMOST ENTIRELY FROM THE EXECUTABLE.
;IT IS POORLY CODED, BUT MOSTLY UNCHANGED FOR COMPATIBILITY
;WITH MAIN FUNCTION.
;GetInstanceFromStringGlobal should generally be used instead.
;GetInstanceFromString should be used during initialization only.
;GetInstanceFromStringGlobal may not return instances if it
;is called too early during loading (instances not initialized).

push esi
push ecx
push edx

     mov esi,eax
     push 16
     call [YAPAXI]
     add esp,4
     test eax,eax
	  jz .notok
     push [GETINST_ADDR1]
     mov ecx,eax
     call [GETINST_PROC1]
     jmp .isok
     .notok:
	 xor eax,eax
     .isok:

     push eax
     push 0
     sub esp,8
     mov ecx,esp
     push esi
     call [GETINST_PROC2]
     mov eax,[GETINST_ADDR2]
     mov edx,[eax]
     mov ecx,[edx+4]
     call [GETINST_PROC3]      ;takes 4 parameters (push eax; push 0; sub esp,8)

pop edx
pop ecx
pop esi
ret

align 16
;============================================================
GetDDrawObject: 	;returns in eax (should be NULL if none)
;============================================================
push ecx
     mov ecx,[DDRAW_OBJ_BASE]
     mov eax,[ecx+4]
     test eax,eax
	  jnz .d3dfound
     mov eax,[ecx]			;(software mode otherwise)
     .d3dfound:
pop ecx
ret

align 16
;============================================================
GetDDSurfaceStructSize:        ;returns in eax (should be NULL if none)
;============================================================
;PRESERVES ecx,edx
push ecx
     mov ecx,[DDRAW_OBJ_BASE]
     mov eax,sizeof.DDSURFACEDESC2
     cmp DWORD [ecx+4],0
	  jnz .d3dfound
     mov eax,sizeof.DDSURFACEDESC		       ;(software mode otherwise)
     .d3dfound:
pop ecx
ret

align 16
;============================================================
CompareStrings: 	;ecx=string1, edx=string2  ;IGNORES CASE
;============================================================
			;returns FALSE if not same, edx and ecx point to null-terminators
.compareloop:		;edi points to dest.'s null-terminator (BYTE [edi] = 0)
     mov al,[ecx]
     mov ah,[edx]	       ;transfer/append string
     CHAR_TOLOWERCASE al
     CHAR_TOLOWERCASE ah
     cmp ah,al
	 jnz .stringsnotequal
     test al,al
	  jz .stringsequal
     inc ecx
     inc edx
     jmp .compareloop
.stringsequal:
     mov eax,TRUE
ret

.stringsnotequal:
     xor eax,eax
ret

align 16
;============================================================
TransferString: 	;esi=source, edi=dest., destroyed: esi,edi, al
;============================================================
.transferloop:		;edi points to dest.'s null-terminator (BYTE [edi] = 0)
     mov al,[esi]
     mov [edi],al	;transfer/append string
     test al,al
	  jz .exitloop
     inc esi
     inc edi
     jmp .transferloop
.exitloop:
ret

align 16
;============================================================
DisableInGameMvtKeys:
;============================================================
push edx
push eax
     mov edx,[KEY_LIST_ADDR]
     xor eax,eax
     mov [edx+4+(5*8)],eax    ;jump
     mov [edx+4+(10*8)],eax   ;run
     mov [edx+4+(11*8)],eax   ;walk
     mov [edx+4+(12*8)],eax   ;backward
     mov [edx+4+(13*8)],eax   ;left
     mov [edx+4+(14*8)],eax   ;right
pop eax
pop edx
ret

align 16
;============================================================
DisableInGameKeys:	      ;includes mouse clicks
;============================================================
push eax
push edx
push esi

     mov edx,[KEY_LIST_ADDR]
     lea esi,[edx+15*8]
     add edx,4
     xor eax,eax

     .disableloop:
     mov [edx],eax
     add edx,8
     cmp edx,esi
	 jb .disableloop

pop esi
pop edx
pop eax
ret

align 16
;============================================================
EnableInGameKeys:	       ;includes mouse clicks
;============================================================
push eax
push edx
push esi

     mov edx,[KEY_LIST_ADDR]
     mov esi,DefaultKeyBOOLS
     add edx,4

     .enableloop:
     mov eax,[esi]
     mov [edx],eax
     add edx,8
     add esi,4
     cmp esi,DefaultKeyBOOLS_END
	 jb .enableloop

pop esi
pop edx
pop eax
ret

align 16
;============================================================
RANDOMIZE_RANDOM_PROC:
;============================================================
invoke time,0
invoke srand,eax
add esp,8
ret

align 16
;============================================================
GetRandomNumber:	;[esp+4] = below or equal to this number and not zero
;============================================================
invoke rand
mov ecx,[esp+4]
xor edx,edx
test ecx,ecx
     jz .dividebyzero
div ecx
mov eax,edx		;remainder
add eax,1
ret 4			;returns number in eax (0 if param was null)

.dividebyzero:
xor eax,eax
ret 4

align 16
;============================================================
IsLevelSavedGame:	;returns 0 if false, non-zero if true
;============================================================
;NOTE: If a Level Restart was performed before last load,
;      automatically NOT a saved game.
push edi
push ecx
push esi

	or ecx,0FFFFFFFFh
	mov edi,[LEVELNAME_OBJ]
	add edi,20
	xor eax,eax
	repnz scasb

	cmp DWORD [edi-5],'.SCN'
	    jz .notsavedgame
	cmp DWORD [edi-5],'.scn'
	    jz .notsavedgame
	cmp [Restart_Exit_Level_BOOL],1
	    jz .notsavedgame
	or al,1
	.notsavedgame:

pop esi
pop ecx
pop edi
ret

align 16
;============================================================
IsLevelSavedGameAbsolute:	;returns 0 if false, non-zero if true
;============================================================
;NOTE: Does not take into account level restarts (source *.scn
;      or savegame name only)
push edi
push ecx
push esi

	or ecx,0FFFFFFFFh
	mov edi,[LEVELNAME_OBJ]
	add edi,20
	xor eax,eax
	repnz scasb

	cmp DWORD [edi-5],'.SCN'
	    jz .notsavedgameabs
	cmp DWORD [edi-5],'.scn'
	    jz .notsavedgameabs
	or al,1
	.notsavedgameabs:

pop esi
pop ecx
pop edi
ret

align 16
;============================================================
	GetPrivateProfileFloat	dd	GetINIFloat
;============================================================
GetINIFloat:	;returns 32-bit float in eax (null if doesn't exist or if 0)
;============================================================
;[esp+4]=points to section name
;[esp+8]=points to key name
;[esp+12]=default string if none found (pointer)
;[esp+16]=points to INI filename
push ebp
mov ebp,esp
sub esp,104

	lea edx,[ebp-100]
invoke GetPrivateProfileString,[ebp+8],[ebp+12],[ebp+16],edx,\
		100,[ebp+20]

	lea edx,[ebp-104]
	lea eax,[ebp-100]

	cmp BYTE [eax],';'
	    jnz NoEmptyFloatEntry2
	    mov eax,[ebp+16]
	    jmp NoEmptyFloatEntry
	NoEmptyFloatEntry2:

	cmp BYTE [eax],0
	    jnz NoEmptyFloatEntry
	    mov eax,[ebp+16]
	NoEmptyFloatEntry:

invoke sscanf,eax,_float,edx
	add esp,12

	pop eax
mov esp,ebp
pop ebp
ret 16

align 16
;============================================================
;Quaternion multiplication (eax=q1, ecx=q2, edx=result)
;============================================================

virtual at eax
	w1	dd	?
	x1	dd	?
	y1	dd	?
	z1	dd	?
end virtual

virtual at ecx
	w2	dd	?
	x2	dd	?
	y2	dd	?
	z2	dd	?
end virtual

;(Q1 * Q2).w = (w1w2 - x1x2 - y1y2 - z1z2)
;(Q1 * Q2).x = (w1x2 + x1w2 + y1z2 - z1y2)
;(Q1 * Q2).y = (w1y2 - x1z2 + y1w2 + z1x2)
;(Q1 * Q2).z = (w1z2 + x1y2 - y1x2 + z1w2)

QuaternionMul_STD:	  ;eax=q1 , ecx=q2, edx=new quat
;------------- w
	fld [w1]
	fmul [w2]

	fld [x1]
	fmul [x2]
	fsubp st1,st0

	fld [y1]
	fmul [y2]
	fsubp st1,st0

	fld [z1]
	fmul [z2]
	fsubp st1,st0
	fstp DWORD [edx]

;------------- x
	fld [w1]
	fmul [x2]

	fld [x1]
	fmul [w2]
	faddp st1,st0

	fld [y1]
	fmul [z2]
	faddp st1,st0

	fld [z1]
	fmul [y2]
	fsubp st1,st0
	fstp DWORD [edx+4]

;------------- y
	fld [w1]
	fmul [y2]

	fld [x1]
	fmul [z2]
	fsubp st1,st0

	fld [y1]
	fmul [w2]
	faddp st1,st0

	fld [z1]
	fmul [x2]
	faddp st1,st0
	fstp DWORD [edx+8]

;------------- z
	fld [w1]
	fmul [z2]

	fld [x1]
	fmul [y2]
	faddp st1,st0

	fld [y1]
	fmul [x2]
	fsubp st1,st0

	fld [z1]
	fmul [w2]
	faddp st1,st0
	fstp DWORD [edx+12]

ret


align 16
;Note: I don't know if you need to normalize locally-generated quaternions. I believe
;not, but I'm leaving it here nevertheless because I'm no expert and I don't want
;to risk messing up any existing features.
;============================================================
ToLocalQuat_AxisX_STD:	  ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 1.00000 , axis.y=0.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	fld st0
	fsin
	fst DWORD [edx+4]	;x
	fmul st0,st0

	fxch
	fcos
	fst DWORD [edx] 	;w
	fmul st0,st0
	faddp st1,st0

	fsqrt			;magnitude
	xor eax,eax
	fdivr [float_one]	;1/magnitude
	fld st0

	fmul DWORD [edx]	;normalize w
	      mov [edx+8],eax	;y
	fstp DWORD [edx]
	fmul DWORD [edx+4]	;normalize x
	      mov [edx+12],eax	;z
	fstp DWORD [edx+4]

ret

align 16
;============================================================
ToLocalQuat_AxisY_STD:	  ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 0.00000 , axis.y=1.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	fld st0
	fsin
	fst DWORD [edx+8]	;y
	fmul st0,st0

	fxch
	fcos
	fst DWORD [edx] 	;w
	fmul st0,st0
	faddp st1,st0

	fsqrt			;magnitude
	xor eax,eax
	fdivr [float_one]	;1/magnitude
	fld st0

	fmul DWORD [edx]	;normalize w
	      mov [edx+4],eax	;x
	fstp DWORD [edx]
	fmul DWORD [edx+8]	;normalize y
	      mov [edx+12],eax	;z
	fstp DWORD [edx+8]

ret

align 16
;============================================================
ToLocalQuat_AxisZ_STD:	  ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 0.00000 , axis.y=0.0000, axis.z=1.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	fld st0
	fsin
	fst DWORD [edx+12]	 ;z
	fmul st0,st0

	fxch
	fcos
	fst DWORD [edx] 	;w
	fmul st0,st0
	faddp st1,st0

	fsqrt			;magnitude
	xor eax,eax
	fdivr [float_one]	;1/magnitude
	fld st0

	fmul DWORD [edx]	;normalize w
	      mov [edx+8],eax	;y
	fstp DWORD [edx]
	fmul DWORD [edx+12]	;normalize z
	      mov [edx+4],eax	;x
	fstp DWORD [edx+12]

ret

align 16
;============================================================
ToLocalQuat_AxisX_NoNorm_STD:	 ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 1.00000 , axis.y=0.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+8],eax
	mov DWORD [edx+12],eax
	fstp DWORD [edx]
	fstp DWORD [edx+4]

ret

align 16
;============================================================
ToLocalQuat_AxisMinusX_NoNorm_STD:    ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = -1.00000 , axis.y=0.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+8],eax
	mov DWORD [edx+12],eax
	fstp DWORD [edx]
	fchs
	fstp DWORD [edx+4]

ret

align 16
;============================================================
ToLocalQuat_AxisY_NoNorm_STD:	 ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 0.00000 , axis.y=1.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+4],eax
	mov DWORD [edx+12],eax
	fstp DWORD [edx]
	fstp DWORD [edx+8]

ret

align 16
;============================================================
ToLocalQuat_AxisMinusY_NoNorm_STD:
;============================================================
				;axis.x = 0.00000 , axis.y=-1.0000, axis.z=0.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+4],eax
	mov DWORD [edx+12],eax
	fstp DWORD [edx]
	fchs
	fstp DWORD [edx+8]

ret

align 16
;============================================================
ToLocalQuat_AxisZ_NoNorm_STD:	 ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 0.00000 , axis.y=0.0000, axis.z=1.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+4],eax
	mov DWORD [edx+8],eax
	fstp DWORD [edx]
	fstp DWORD [edx+12]

ret

align 16
;============================================================
ToLocalQuat_AxisMinusZ_NoNorm_STD:    ;ecx=pointer to radian angle to rotate by, edx=local_rotation pointer wxyz
;============================================================
				;axis.x = 0.00000 , axis.y=0.0000, axis.z=-1.00000
	fld DWORD [ecx]
	fmul [float_0_5]
	xor eax,eax
	fsincos
	mov DWORD [edx+4],eax
	mov DWORD [edx+8],eax
	fstp DWORD [edx]
	fchs
	fstp DWORD [edx+12]

ret

align 16
;============================================================
QuatToAxisAngle_STD:	    ;eax = w,x,y,z quat ;ecx = angle+axis (angle in radians, axisx, axisy, axisz)
;============================================================
sub esp,16
	mov edx,[eax]
	mov [esp],edx
	mov edx,[eax+4]
	mov [esp+4],edx
	mov edx,[eax+8]
	mov [esp+8],edx
	mov edx,[eax+12]
	mov [esp+12],edx
	mov eax,esp
	call NormalizeQuat

	fld DWORD [eax] 	;cos
	fmul st0,st0
	fsubr [float_one]
	fsqrt			;sin = sqrt(1.0-cos^2)
	fld st0
	fld DWORD [eax]
	fpatan			;acos = tan_-1 (sin/cos)
	fmul [float_two]	;angle = acos*2
	fstp DWORD [ecx]

	fld st0
	fabs
	fcomp [float_0_0005]	 ;sin now in st0
	push eax
	fstsw ax
	test eax,100h		;sin < 0.0005 ?
	     jz sinnotbelow0_0005
	     fstp st0
	     fld1		;if so, sin = 1.000000
	     sinnotbelow0_0005:
	pop eax

	fdivr [float_one]	;1/sin

	fld DWORD [eax+4]	;x
	fmul st0,st1		;x/sin
	fstp DWORD [ecx+4]	;axisx

	fld DWORD [eax+8]	;y
	fmul st0,st1		;y/sin
	fstp DWORD [ecx+8]	;axisy

	fld DWORD [eax+12]	;z
	fmul st0,st1		;z/sin
	fstp DWORD [ecx+12]	;axisz

	fstp st0

add esp,16
ret

align 16
;============================================================
NormalizeQuat_STD:	    ;eax = quat (w,x,y,z)
;============================================================
	fld DWORD [eax]
	fmul st0,st0
	fld DWORD [eax+4]
	fmul st0,st0
	fld DWORD [eax+8]
	fmul st0,st0
	fld DWORD [eax+12]
	fmul st0,st0
	faddp
	faddp
	faddp
	fsqrt

	fdivr [float_one]

	fld st0
	fmul DWORD [eax]
	fstp DWORD [eax]
	fld st0
	fmul DWORD [eax+4]
	fstp DWORD [eax+4]
	fld st0
	fmul DWORD [eax+8]
	fstp DWORD [eax+8]

	fmul DWORD [eax+12]
	fstp DWORD [eax+12]

ret

align 16
;============================================================
CONVERT_TO_FACING_PROC_2A_STD:	   ;[esp+4] = destvar, [esp+8]=param2, source quat=ecx
;============================================================
;
;CONVERT_TO_FACING_PROC:
; - quaternion to convert in ecx (w = c0, x = c4, y = c8, z = cC)
; - param2 put into eax in original function ([eax+00h] = a0, [eax+04h] = a4, [eax+08h] = a8)
; - final values put into destvar (first parameter)
; - original function uses local mem to store data (4 DWORDs at esp)
; - qword constant or var stored in EXE used once (referred to as q); should be 0.500000
;
; -> new function uses edx for param2 and eax for destvar, because destvar must
;    be returned in eax anyway


virtual at ecx
	c0	dd	?
	c4	dd	?
	c8	dd	?
	cC	dd	?
end virtual

virtual at edx
	a0	dd	?
	a4	dd	?
	a8	dd	?
end virtual

	fld DWORD [c0]
	fmul st0,st0
	mov eax,[CONVERTTOFACING_VAR1]
	fsub qword [eax]	 ;q (var or constant)

	fld [c0]
	fmul [cC]
	fld [c0]
	fmul [c8]
	fld [c4]
	fmul [cC]
	fld [c0]
	fmul [c4]

	fld [c8]
	fmul [cC]		;st2 when full stack

	fld [c4]
	fmul [c8]
	mov edx,[esp+8]
	mov eax,[esp+4]
	fstp DWORD [esp-4]     ;(c4*c8)

;DWORD [destvar+00h] = [eax] = [esp+04h] (originally)
;          = ((c4^2+(c0^2-q))*a0 + (c4*c8-c0*cC)*a4 + (c0*c8+c4*cC)*a8)*2

	fld [c4]
	fmul st0,st0
	fadd st0,st6
	fmul [a0]

	fld DWORD [esp-4]
	fsub st0,st6
	fmul [a4]
	faddp st1,st0

	fld st4
	fadd st0,st4
	fmul [a8]
	faddp st1,st0
	fadd st0,st0
	fstp DWORD [eax]

;DWORD [destvar+04h] = [eax+4] = [esp+08h] (originally)
;          = ((c8^2+(c0^2-q))*a4 + (c8*cC-c0*c4)*a8 + (c0*cC+c4*c8)*a0)*2

	fld [c8]
	fmul st0,st0
	fadd st0,st6
	fmul [a4]

	fld st1
	fsub st0,st3
	fmul [a8]
	faddp st1,st0

	fld DWORD [esp-4]
	fadd st0,st6
	fmul [a0]
	faddp st1,st0
	fadd st0,st0
	fstp DWORD [eax+4]

;DWORD [destvar+08h] = [eax+8] = [esp+0Ch] (originally)
;          = ((cC^2+(c0^2-q))*a8 + (c4*cC-c0*c8)*a0 + (c0*c4+c8*cC)*a4)*2

	faddp st1,st0		;c0*c4+c8*cC
	fmul [a4]
	fxch st2
	fsubp st1,st0		;c4*cC-c0*c8
	fmul [a0]
	faddp st1,st0

	fld [cC]
	fmul st0,st0
	fadd st0,st3
	fmul [a8]

	fxch st2
	fstp st0
	faddp st1,st0
	fadd st0,st0
	fxch st1
	fstp st0
	fstp DWORD [eax+8]


ret 8

align 16
;============================================================
CONVERT_TO_FACING_PROC_010_STD:      ;[esp+4] = destvar. (eax), [esp+8] = (unused) , ecx = quat
;============================================================
;This is used when the first param points to:
; DWORD 0.000000, DWORD 1.000000, DWORD 0.000000
;This is standard axis data (I believe), used for LOC cheat facing values,
;FLY cheat and others. Parameters are preserved for future implementation
;and compatibility.

;[destvar+00h] = ((c4^2+(c0^2-q))*a0 + (c4*c8-c0*cC)*a4 + (c0*c8+c4*cC)*a8)*2
;[destvar+04h] = ((c8^2+(c0^2-q))*a4 + (c8*cC-c0*c4)*a8 + (c0*cC+c4*c8)*a0)*2
;[destvar+08h] = ((cC^2+(c0^2-q))*a8 + (c4*cC-c0*c8)*a0 + (c0*c4+c8*cC)*a4)*2
;Where a0 = 0.000000, a4 = 1.000000, a8 = 0.000000
;Result (faster!):
;[destvar+00h] = (c4*c8-c0*cC)*2     (X facing value)
;[destvar+04h] = (c8^2+(c0^2-q))*2   (Y facing value)
;[destvar+08h] = (c0*c4+c8*cC)*2     (Z facing value)

	mov eax,[esp+4]
	fld [cC]
	fld [c8]
	fld [c4]
	fld [c0]

	fld st0
	fmul st0,st0	     ;c0^2
	fld st3 	     ;c8
	fmul st0,st0	     ;c8^2
	fxch st1
	fsub [float_0_5]
	fxch st1
	faddp st1,st0
	fadd st0,st0	     ;*2
	fstp DWORD [eax+4]

	fld st1 	     ;c4
	fmul st0,st3	     ;c4*c8
	fld st1 	     ;c0
	fmul st0,st5	     ;c0*cC
	fxch st1
	fsubrp st1,st0	     ;c4*c8-c0*cC
	fadd st0,st0	     ;*2
	fstp DWORD [eax]

	fmulp st1,st0	     ;c0*c4
	fxch st2
	fmulp st1,st0	     ;c8*cC
	fxch st1
	faddp st1,st0	     ;c0*c4+c8*cC
	fadd st0,st0	     ;*2
	fstp DWORD [eax+8]

ret 8

align 16
;============================================================
GetInstancePhysicsVelocity:  ;includes rotation speed
;============================================================
;unlike EXE, returns TRUE if object is active and NULL if
;default values were returned.

     mov edx,[GETOBJPHYSICSOBJADDR]
     push DWORD [esp+8]
     mov ecx,[edx]
     call [GETOBJPHYSICSINDEXPROC]
     test eax,eax
	  jl GetInstancePhysicsVelocity_OBJStill

     cmp [EXEID],'DEMO'
	 jnz GetInstancePhysicsVelocityFV_2
     mov ecx,eax		   ;for DEMO EXE
     shl ecx,5
     add ecx,eax
     lea eax,[eax+8*ecx]
     shl eax,4
     jmp GetInstancePhysicsVelocityDEMO_2
GetInstancePhysicsVelocityFV_2:
     mov ecx,eax		   ;for full version EXEs
     shl ecx,6
     add ecx,eax
     lea edx,[eax+2*ecx]
     lea eax,[eax+8*edx]
     shl eax,2
GetInstancePhysicsVelocityDEMO_2:
     add eax,[OBJPHYSICSBASEADDR]

     push edi
     push esi
     mov edi,[esp+12]
     lea esi,[eax+4]
     mov ecx,7
     CurrentInstanceVelocitySetLoop:
     mov eax,[esi]
     mov [edi],eax
     add esi,0Ch
     add edi,4				;7 values are at 0Ch intervals
     loop CurrentInstanceVelocitySetLoop
     pop esi
     pop edi

mov eax,1
ret 8

GetInstancePhysicsVelocity_OBJStill:
     push edi
     mov edi,[esp+8]
     mov eax,0.000000		  ;same as EXE, but assigns 0.000000 instead of
     mov ecx,7			  ;1.000000 to Wquat velocity value *
     rep stosd
     mov edx,[esp+8]
     pop edi
     ;mov DWORD [edx+12],1.000000
xor eax,eax
ret 8

;* Theoretically, the value 1.000000 should be the correct one,
;because it is part of the zero-rotation local quaternion (includingcosine of 0). However,
;due to the way the values are set when an object is first moved
;from a frozen state, it must be set to 0.000000 to avoid giving
;the object an initial rotation, despite possibly inexact math.

align 16
;============================================================
SetInstancePhysicsUnfreezeOnly:
;============================================================
     mov eax,[esp+04h]
     push [SETOBJPHYSICSOBJPARAM]
     mov edx,[ecx]
     push 0
     push eax
     call DWORD [edx+18h]	     ;unfreeze object/enable object physics

ret 8

align 16
;============================================================
SetInstancePhysicsVelocity:  ;includes rotation speed
;============================================================
;This is same as the [+44h] method used by ActionType 10 to set [X,Y,Z]
;velocity, except will also set quaternion rotation velocity.
;[esp+4] = instance
;[esp+8] = address of 7 velocity values
;ecx = physics sub-instance of instance [[instance+28h]+0Ch]
;Note: The DEMO executable finds the base address of the velocity
;      using different calculations than the full-version 1.1 EXEs.

push edi
push esi

     mov edi,[esp+0Ch]
     push [SETOBJPHYSICSOBJPARAM]
     mov edx,[ecx]
     push 0
     push edi
     call DWORD [edx+18h]	     ;unfreeze object/enable object physics

     mov edx,[GETOBJPHYSICSOBJADDR]
     push edi
     mov ecx,[edx]
     call [GETOBJPHYSICSINDEXPROC]   ;ecx = sub-instance, already set

     cmp [EXEID],'DEMO'
	 jnz GetInstancePhysicsVelocityFV
     mov ecx,eax		   ;for DEMO EXE
     shl ecx,5
     add ecx,eax
     lea eax,[eax+8*ecx]
     shl eax,4
     jmp GetInstancePhysicsVelocityDEMO
GetInstancePhysicsVelocityFV:
     mov ecx,eax		   ;for full version EXEs
     shl ecx,6
     add ecx,eax
     lea edx,[eax+2*ecx]
     lea eax,[eax+8*edx]
     shl eax,2
GetInstancePhysicsVelocityDEMO:
     add eax,[OBJPHYSICSBASEADDR]

     mov esi,[esp+16]
     lea edi,[eax+4]
     mov ecx,7
     CurrentInstanceVelocityGetLoop:
     mov eax,[esi]
     mov [edi],eax
     add esi,4
     add edi,0Ch			  ;7 values are at 0Ch intervals
     loop CurrentInstanceVelocityGetLoop

pop esi
pop edi
ret 8

align 16
;============================================================
VectorDotProduct_STD:	;performs: (vector1 -dot product- vector2)/
;============================================================
;ecx = pointer to source vector [x,y,z]
;edx = pointer to dest./directional vector [x,y,z]
;eax = address of 1 variable to contain vector value (signed, float)

     fld DWORD [edx]	    ;begin dot product
     fmul DWORD [ecx]
     fld DWORD [edx+4]
     fmul DWORD [ecx+4]
     fld DWORD [edx+8]
     fmul DWORD [ecx+8]
     faddp
     faddp		    ;end dot product
     fstp DWORD [eax]

ret

align 16
;============================================================
FDConvAnglesToQuat_STD:        ;converts two rotation angles to a quat
;============================================================
;ecx = pointer to 2 angle rotation values already converted to radians (x,z) (y calculated from x)
;      x component = angle with positive side of y axis (will be negated)
;      z component = angle with x-y plane
;[esp+4] = pointer to 4 DWORD struc to receive (implicitely normalized) quat wxyz

sub esp,36

      mov eax,[ecx]
      mov edx,esp
      mov [esp+32],eax
      add ecx,4
      call ToLocalQuat_AxisX_NoNorm   ;calculate z rotation quat (around x axis)

      fld DWORD [esp+32]
      lea ecx,[esp+32]
      fchs
      lea edx,[esp+16]
      fstp DWORD [ecx]
      call ToLocalQuat_AxisZ_NoNorm   ;calculate x rotation quat (around z axis)

      mov ecx,esp
      lea eax,[esp+16]
      mov edx,[esp+36+4]
      call QuaternionMul

add esp,36

ret 4

align 16
;============================================================
FacingRad2CompToVector_Sub_STD:        ;assumes dest. vector size = 1
;============================================================
;ecx = pointer to 2 angle rotation values already converted to radians (x,z) (y calculated from x)
;      x component = angle with positive side of y axis
;      z component = angle with x-y plane
;edx = pointer to 3 position values [x,y,z] to add vector to
;eax = pointer to distance multiplier
;note: input angles are based on but do not follow the standard facing values. The X component
;      is NOT limited to the -1.000000 to +1.0000000 interval here, which eliminates the need
;      for the y angle (semi-indirectly).
;
;(note2 : fcos of x = same as fsin of y facing for standard facing values)

     fld DWORD [ecx+4]	      ;Z facing
     fsincos		      ;st0 = cos value (vector on X-Y plane), st1 = sin value (Z vector)
     ;fabs
     fmul DWORD [eax]	      ;distance multiplier for vector on X-Y plane from z angle

     fld DWORD [ecx]	      ;X facing
     fsincos

     fld st2		      ;(vector on X-Y plane from fcos of z angle w/distance mul.)*fcos of x
     fmulp
     fsubr DWORD [edx+4]      ;sub from y comp. of vector
     fstp DWORD [edx+4]

     fmulp		      ;(vector on X-Y plane from fcos of z angle w/distance mul.)*fsin of x
     fsubr DWORD [edx]	      ;sub from x comp. of vector
     fstp DWORD [edx]

     fmul DWORD [eax]	      ;distance multiplier*(fsin of z angle)
     fsubr DWORD [edx+8]      ;sub from z comp. of vector
     fstp DWORD [edx+8]

ret

align 16
;============================================================
ToLocalQuat_NoNorm_STD:
;============================================================
;ecx=pointer to 4 DWORD struc: angle (rad), x axis, y axis, z axis values
;edx=local_rotation pointer wxyz

     fld DWORD [ecx]
     fmul [float_0_5]
     fsincos
     fstp DWORD [edx]

     fld st0
     fmul DWORD [ecx+4]
     fstp DWORD [edx+4]
     fld st0
     fmul DWORD [ecx+8]
     fstp DWORD [edx+8]

     fmul DWORD [ecx+12]
     fstp DWORD [edx+12]

ret

align 16
;============================================================
VectorNormalize_STD:		 ;eax = vector (must not be null)
;============================================================
;not optimized
;eax unchanged

     fld DWORD [eax]
     fmul st0,st0
     fld DWORD [eax+4]
     fmul st0,st0
     fld DWORD [eax+8]
     fmul st0,st0
     faddp st1,st0
     faddp st1,st0
     fsqrt

     fdivr [float_one]

     fld DWORD [eax]
     fmul st0,st1
     fstp DWORD [eax]

     fld DWORD [eax+4]
     fmul st0,st1
     fstp DWORD [eax+4]

     fld DWORD [eax+8]
     fmulp st1,st0
     fstp DWORD [eax+8]

ret

align 16
;============================================================
ResetCCameraPlayerSources:
;============================================================
       mov edx,[LocationADDR]
       mov eax,[PLAYER_OBJ_INVUL_ADDR]
       mov edx,[edx]
       mov eax,[eax]
       mov ecx,[edx+14h]		 ;get CCamera object
       mov [ecx+50h],eax
       mov [ecx+54h],eax
ret

align 16
;============================================================
SpanResize24BitColor:
;============================================================
;[esp+4] = color to convert
;[esp+8] = new color span (maximum color value)
;destroys NO registers except eax (= result)
;NOT OPTIMIZED
;This function will convert a color from a 00FFFFFFh 256-color value component
;to a 00RRGGBBh RR/GG/BB-color value format, where each color component is less or equal to
;0FFh. If all color components are the same, this will cause the
;color simply to diminish in intensity and keep its original proportionality.
;[esp+8] represents the maximum possible color value.

push ecx
push edx

       mov ecx,[esp+12]
       mov edx,[esp+16]
       call SpanResize8BitColor
       mov eax,ecx

       mov ecx,[esp+12]
       mov edx,[esp+16]
       shr ecx,8
       shr edx,8
       call SpanResize8BitColor
       shl ecx,8
       or eax,ecx

       mov ecx,[esp+12]
       mov edx,[esp+16]
       shr ecx,16
       shr edx,16
       call SpanResize8BitColor
       shl ecx,16
       or eax,ecx

pop edx
pop ecx
ret 8

align 16
;============================================================
SpanResize8BitColor:	;cl = color to convert, dl = max span -> result in ecx (cl)
;============================================================
;destroyed: ecx, edx

       and ecx,0FFh
       and edx,0FFh
       push ecx
       push edx
       fild DWORD [esp+4]		;cl/255 = new/maxspan -> new = (cl/255)*maxspan)
       fmul [float_1div255]
       fild DWORD [esp]
       fmulp
       fistp DWORD [esp]
       add esp,8
       mov cl,[esp-8]

ret

align 16
;============================================================
;Limited Quaternion multiplication (eax=q1, ecx=q2, edx=result)
;============================================================
;supposes second quaternion is rotation around x axis only (y2 and z2 null)
;eax = source quat (all valid)
;ecx = around-x quat
;w2 = non-zero
;x2 = non-zero
;y2 = null
;z2 = null

;(Q1 * Q2).w = (w1w2 - x1x2)
;(Q1 * Q2).x = (w1x2 + x1w2)
;(Q1 * Q2).y = (y1w2 + z1x2)
;(Q1 * Q2).z = (z1w2 - y1x2)

QuaternionMulLocalQ2AxisX_STD:	      ;eax=q1 , ecx=q2, edx=new quat
;------------- w

	fld [w1]
	fmul [w2]

	fld [x1]
	fmul [x2]
	fsubp st1,st0

	fstp DWORD [edx]

;------------- x
	fld [w1]
	fmul [x2]

	fld [x1]
	fmul [w2]
	faddp st1,st0

	fstp DWORD [edx+4]

;------------- y

	fld [y1]
	fmul [w2]

	fld [z1]
	fmul [x2]
	faddp st1,st0
	fstp DWORD [edx+8]

;------------- z

	fld [z1]
	fmul [w2]

	fld [y1]
	fmul [x2]

	fsubp st1,st0
	fstp DWORD [edx+12]

ret

align 16
;============================================================
CCameraTeleport:	;edx = pointer to 3 location values (y,x,z) and 4 quat values (w,x,y,z)
;============================================================
;destroyed: eax,ecx,edx
;If members DWORD [CCamera+50h] and DWORD [CCamera+54h] are CPlayerPriv objects,
;the player object will be teleported along with the camera.

       mov eax,[LocationADDR]
       mov eax,[eax]
       mov ecx,[eax+14h]		 ;get CCamera object

       push 0
       mov eax,[ecx]
       push edx
       call DWORD [eax+ICInstance.Teleport]    ;set new camera position/orientation

ret

align 16
;============================================================
GetInstanceFromStringGlobal:	;eax = pointer to string, returns instance in eax or null if not found
;============================================================
;should not be used in performance-critical areas where
;the second part of this function is to be called repeatedly
;on the same string ID (from the same string)

;GetInstanceFromStringGlobal may not return instances if it
;is called too early during loading (instances not initialized).

       test eax,eax
	    jz GetInstanceFromStringGlobal_END

	invoke GETTEMPINST_PROC,eax,0,1    ;returns string ID (case-sensitive, unique)
							  ;params 0 and 1 are standard -> IN SOME CASES, PARAM2 MUST EQUAL LENGTH OF STRING (such as when master instances????)
	add esp,12
	mov edx,[LocationADDR]
	push eax
	mov ecx,[edx]
	call [GETOBJFROMSTRINGIDPROC]	    ;this fetches the string ID's associated instance, if existing

GetInstanceFromStringGlobal_END:
ret

align 16
;============================================================
GetInstanceFromStringGlobal_CBD:    ;eax = pointer to string, returns instance in eax or null if not found
;============================================================
;also check in CBackDrop buffer
push ebx
       test eax,eax
	    jz GetInstanceFromStringGlobal_CBD_END

	invoke GETTEMPINST_PROC,eax,0,1    ;returns string ID (case-sensitive, unique)
							  ;params 0 and 1 are standard -> IN SOME CASES, PARAM2 MUST EQUAL LENGTH OF STRING (such as when master instances????)
	add esp,12
	mov ebx,eax

	mov edx,[LocationADDR]
	push eax
	mov ecx,[edx]
	call [GETOBJFROMSTRINGIDPROC]	    ;this fetches the string ID's associated instance, if existing
	test eax,eax
	     jnz GetInstanceFromStringGlobal_CBD_END

	mov eax,ebx
	call CBACKDROPFINDINSTANCEFROMID

GetInstanceFromStringGlobal_CBD_END:
pop ebx
ret

align 16
;============================================================
GetCAnimalInstanceCount:      ;returns count in eax (for current level)
;============================================================
	mov eax,[CAnimal_Intance_Count]
ret

align 16
;============================================================
VectorMagnitude_STD:	      ;eax = pointer to vector, edx = pointer for result
;============================================================

     fld DWORD [eax]
     fmul st0,st0
     fld DWORD [eax+4]
     fmul st0,st0
     fld DWORD [eax+8]
     fmul st0,st0
     faddp st1,st0
     faddp st1,st0
     fsqrt
     fstp DWORD [edx]

ret


align 16
;============================================================
VectorToQuaternion_STD:    ;ecx = pointer to 3 XYZ source vector values, edx = destination for WXYZ quat (returned in eax)
;============================================================
push esi
push edi
	mov esi,ecx
	mov edi,edx
sub esp,36

    ;calculate horizontal rotation quaternion (x-y)
	fld DWORD [esi+0]	;x
	fld DWORD [esi+4]	;y (note: y is base axis)
	fpatan			;arctan(x/y)    (let processor handle problems)
				;if 0/0, return value makes no difference (will point upward - z only)
	mov ecx,esp
	lea edx,[esp+4] 	;for local quat
	fstp DWORD [ecx]
	call [PTRToLocalQuat_AxisMinusZ_NoNorm]   ;generate horizontal rotation quaternion

    ;calculate vertical rotation quaternion (z)  - SIGN ERROR
	fld DWORD [esi+8]	;z vector component
	fld DWORD [esi+4]
	fmul st0,st0
	fld DWORD [esi+0]
	fmul st0,st0
	faddp st1,st0
	fsqrt			;sqrt(x^2+y^2)
	fpatan			;arctan(z/sqrt(x^2+y^2))

	mov ecx,esp
	lea edx,[esp+4+16]	;for local quat
	fstp DWORD [ecx]
	call [PTRToLocalQuat_AxisX_NoNorm]

	lea ecx,[esp+4+16]
	lea eax,[esp+4]
	mov edx,edi
	call QuaternionMul	;multiply quaternions

	mov eax,edi
add esp,36
pop edi
pop esi
ret

align 16
;============================================================
IsValidInstanceVTable:		   ;ecx = instance's VTable pointer member
;============================================================
;returns non-zero if true (tests approximately)

	mov eax,[RDATA_START_ADDR]
	add eax,4
	cmp ecx,eax
	mov eax,[DATASEC_STARTADDR]
	    jb IsValidInstance_NO    ;is pointer in data section?
	sub eax,4
	cmp ecx,eax
	    jnb IsValidInstance_NO

	mov eax,[ecx]		     ;get first VTable member (should be function address)
	sub eax,00401000h	     ;(same for all EXEs)
	    jb IsValidInstance_NO
	cmp eax,[V_CODE_SIZE]
	    jnb IsValidInstance_NO

	mov eax,[ecx-4] 	     ;get VTable debug info member (-4) (should be in data section)
	cmp eax,[RDATA_START_ADDR]
	    jb IsValidInstance_NO
	cmp eax,[DATASEC_STARTADDR]
	    jnb IsValidInstance_NO

	mov eax,1
ret


IsValidInstance_NO:
	xor eax,eax

ret

align 16
;============================================================
IsValidInstance:     ;param1 = mem pointer, param2 = mem size
;============================================================
;returns non-zero if valid (will use return value of IsValidInstance)

	cmp DWORD [esp+4],0
	    jz IsValidInstance_INVALID

	invoke IsBadWritePtr,[esp+8],[esp+8]	       ;test for pointer validity
	test eax,eax
	    jnz IsValidInstance_INVALID
	invoke IsBadReadPtr,[esp+8],[esp+8]
	test eax,eax
	mov edx,[esp+4]
	     jnz IsValidInstance_INVALID

	mov ecx,[edx]
	call IsValidInstanceVTable	    ;test if first member is indeed a valid VTable (approximately, should always be true)

ret 8

IsValidInstance_INVALID:
	xor eax,eax
ret 8

align 16
;============================================================
FindFirstDWORDListMatch:  ;param1 = list1, param2=list2
;============================================================
;compares two list of DWORDS and returns the first value
;that comes up in both list1 and list2, where list1 values
;have priority. Both lists must be null-terminated (by DWORD 0).
;If no match found, implicitly returns 0.

	mov ecx,[esp+4]
	mov edx,[esp+8]

	cmp DWORD [ecx],0
	    jz FindFirstDWORDListMatch_NOTFOUND
	cmp DWORD [edx],0
	    jz FindFirstDWORDListMatch_NOTFOUND

	FindFirstDWORDListMatch_List1Loop:
	mov eax,[ecx]

	    mov edx,[esp+8]
	    FindFirstDWORDListMatch_List2Loop:
	    cmp eax,[edx]
		jz FindFirstDWORDListMatch_FOUND
	    add edx,4
	    cmp DWORD [edx],0
		jnz FindFirstDWORDListMatch_List2Loop

	add ecx,4
	cmp DWORD [edx],0
	    jnz FindFirstDWORDListMatch_List1Loop

	FindFirstDWORDListMatch_NOTFOUND:
	xor eax,eax
	FindFirstDWORDListMatch_FOUND:
ret 8

align 16
;============================================================
XYZDegreesToQuaternion:   ;converts x,y,z angles to a quat
;============================================================
;ecx = pointer to x angle, y angle, z angle
;edx = pointer to dest. for wxyz quaternion
push ebp
mov ebp,esp
sub esp,64
push esi
push edi
	mov esi,ecx
	mov edi,edx

;create three local rotation quats, one per axis
	lea ecx,[esi]
	lea edx,[ebp-48]
	call [PTRToLocalQuat_AxisX_NoNorm]

	lea ecx,[esi+4]
	lea edx,[ebp-32]
	call [PTRToLocalQuat_AxisY_NoNorm]

	lea ecx,[esi+8]
	lea edx,[ebp-16]
	call [PTRToLocalQuat_AxisZ_NoNorm]


	lea eax,[ebp-16]
	lea ecx,[ebp-32]
	lea edx,[ebp-64]
	call [PTRQuaternionMul]      ;z quat * y quat

	lea eax,[ebp-64]
	lea ecx,[ebp-48]
	lea edx,[edi]
	call [PTRQuaternionMul]      ;(z quat * y quat) * x quat

pop edi
pop esi
mov esp,ebp
pop ebp
ret

align 16
;============================================================
Convert3DPixelTo2DRender:
;============================================================
;ecx = pointer to 3 floats
;edx = pointer for 2 int return values (x and y positions)
;      and 1 float for 1/z value
;NOTE: These calculations are basically copied from the Trespasser
;executable (not within a usable function). They are basically
;"black-box" calculations and implicitly calculate screen positions,
;taking into account screen resolution and size. They depend on a
;call to a CCamera method, which returns values used in a dot-product
;-style calculation.
;this function explicitly returns values that take into account
;screen offset caused by screen size change.
;takes care of screen resolution checks
;returns in eax:
;1 if within screen
;0 if not within screen (or if behind screen)
push ebp
mov ebp,esp
sub esp,(12*4)
push edi
push esi
push ebx

	mov eax,[BaseDDSurfaceADDR]
	mov edi,edx
	mov esi,ecx
	mov ebx,[eax]

       ;get camera object
	mov eax,[LocationADDR]
	mov eax,[eax]
	lea edx,[ebp-(12*4)]
	mov ecx,[eax+14h]
	push edx
	call [GetCam2DRenderValuesProc]

       ;calculate depth of point and check if behind screen
	fld [cam3]
	fmul DWORD [esi]
	fld [cam6]
	fmul DWORD [esi+4]
	fld [cam9]
	fmul DWORD [esi+8]
	faddp st1,st0
	faddp st1,st0
	fadd [cam12]

	fst DWORD [edi+8]

	fcomp [cam2d_depth_check]	;is z below minimum? (typical min = 1.00000)
	fstsw ax
	test ah,01h
	     jnz Convert3DPixelTo2DRender_OUTOFSCREEN

	fld DWORD [edi+8]
	fdivr [float_one]		;get 1/z and store temporarily
	fstp DWORD [edi+8]

       ;calculate y value (height) and check if inside limits
       ;(this is independant of resolution)
	fld [cam2]
	fmul DWORD [esi]
	fld [cam5]
	fmul DWORD [esi+4]
	fld [cam8]
	fmul DWORD [esi+8]
	faddp st1,st0
	faddp st1,st0
	fadd [cam11]

	fmul DWORD [edi+8]	 ;y/z
	fst DWORD [edi+4]

	fcomp [cam2d_minwh_check]
	fstsw ax
	test ah,01h
	     jnz Convert3DPixelTo2DRender_OUTOFSCREEN

	fld DWORD [edi+4]
	fcomp [cam2d_maxwh_check]
	fstsw ax
	test ah,41h
	     jz Convert3DPixelTo2DRender_OUTOFSCREEN

       ;calculate x value (width)
	fld [cam1]
	fmul DWORD [esi]
	fld [cam4]
	fmul DWORD [esi+4]
	fld [cam7]
	fmul DWORD [esi+8]
	faddp st1,st0
	faddp st1,st0
	fadd [cam10]

	fmul DWORD [edi+8]	 ;x/z
	fst DWORD [edi]

	fcomp [cam2d_minwh_check]
	fstsw ax
	test ah,01h
	     jnz Convert3DPixelTo2DRender_OUTOFSCREEN

	fld DWORD [edi]
	fcomp [cam2d_maxwh_check]
	fstsw ax
	test ah,41h
	     jz Convert3DPixelTo2DRender_OUTOFSCREEN

       ;from here on, it is assumed pixel is within screen limits
       ;and get final screen coordinates

	fld DWORD [edi]
	fistp DWORD [edi]
	fld DWORD [edi+4]
	fistp DWORD [edi+4]

	mov ecx,[edi]
	mov edx,[edi+4]

	cmp ecx,0
	    jl Convert3DPixelTo2DRender_OUTOFSCREEN
	cmp edx,0
	    jl Convert3DPixelTo2DRender_OUTOFSCREEN
	cmp ecx,[ebx+8] 	       ;x res
	    jnb Convert3DPixelTo2DRender_OUTOFSCREEN
	cmp edx,[ebx+12]	       ;y res
	    jnb Convert3DPixelTo2DRender_OUTOFSCREEN

	add ecx,[ebx+88h]	       ;x window offset
	add edx,[ebx+8Ch]	       ;y window offset
	mov [edi],ecx
	mov [edi+4],edx

	mov eax,1

pop ebx
pop esi
pop edi
mov esp,ebp
pop ebp
ret

Convert3DPixelTo2DRender_OUTOFSCREEN:
	xor eax,eax
pop ebx
pop esi
pop edi
mov esp,ebp
pop ebp
ret

align 16
;============================================================
IsTresClassInheritor:
;============================================================
;ecx = VTable of instance (inheritor)
;edx = VTable of possibly inherited class
;returns eax = non-zero if is inheritor
push esi

	mov eax,[edx-4]
	mov ecx,[ecx-4]
	mov edx,[eax+0Ch]	;comparative pointer
	mov esi,[ecx+10h]
	mov ecx,[esi+8] 	;inheritance count
	mov esi,[esi+0Ch]	;base inheritance list

	IsTresClassInheritor_loop:
	mov eax,[esi]
	cmp [eax],edx
	    jz IsTresClassInheritor_found
	add esi,4
	dec ecx
	    jnz IsTresClassInheritor_loop
	xor eax,eax

pop esi
ret

IsTresClassInheritor_found:
	mov eax,TRUE

pop esi
ret

align 16
;============================================================
GetHeldWeaponInstance:	   ;ecx = which weapon offset
;============================================================
;returns gun in eax or null if holding nothing or if not gun

	mov eax,[ActiveGunRef]
	mov edx,[eax]
	mov eax,[edx]
	mov edx,[eax+08h]

	cmp ecx,GETWEAPON_SECONDARY		;special case (must check two objects)
	    jz GetHeldWeaponInstance_CheckSecondary

	mov eax,[edx+ecx]
	test eax,eax
	mov edx,[CGUNVTABLE]
	     jz GetHeldWeaponInstance_None
	cmp [eax],edx
	    jnz GetHeldWeaponInstance_None

ret

GetHeldWeaponInstance_CheckSecondary:	  ;two cases to verify when not specifying a location for the secondary weapon
	mov eax,[edx+GETWEAPON_SECONDARYWAIST]
	mov ecx,[CGUNVTABLE]
	test eax,eax
	     jz GetHeldWeaponInstance_CheckSecondary_NotWaist
	cmp [eax],ecx
	    jz GetHeldWeaponInstance_CheckSecondary_Found
	GetHeldWeaponInstance_CheckSecondary_NotWaist:

	mov eax,[edx+GETWEAPON_SECONDARYBACK]
	test eax,eax
	     jz GetHeldWeaponInstance_None
	cmp [eax],ecx
	    jnz GetHeldWeaponInstance_None

GetHeldWeaponInstance_CheckSecondary_Found:
ret

GetHeldWeaponInstance_None:
	xor eax,eax

ret

align 16
;============================================================
GetLoadedBitmapDimensions:   ;ecx = buffer containing entire bitmap file, edx = pointer to dest. values (width, height, bpp)
;============================================================
;returns eax = 1 if OK, or null if unknown structure

	add ecx,sizeof.BITMAPFILEHEADER
	mov eax,[ecx+BITMAPINFOHEADER.biSize]
	cmp eax,sizeof.BITMAPINFOHEADER
	    jz GetLoadedBitmapDimensions_InfoHeader
	cmp eax,sizeof.BITMAPCOREHEADER
	    jz GetLoadedBitmapDimensions_CoreHeader

	xor eax,eax
ret

GetLoadedBitmapDimensions_InfoHeader:
	mov eax,[ecx+BITMAPINFOHEADER.biWidth]
	mov [edx],eax
	mov eax,[ecx+BITMAPINFOHEADER.biHeight]
	mov [edx+4],eax
	xor eax,eax
	mov ax,[ecx+BITMAPINFOHEADER.biBitCount]
	mov [edx+8],eax

	mov eax,TRUE
ret


GetLoadedBitmapDimensions_CoreHeader:
	xor eax,eax
	mov ax,[ecx+BITMAPCOREHEADER.bcWidth]
	mov [edx],eax
	mov ax,[ecx+BITMAPCOREHEADER.bcWidth]
	mov [edx+4],eax
	mov ax,[ecx+BITMAPCOREHEADER.bcBitCount]
	mov [edx+8],eax

	mov eax,TRUE
ret

align 16
;============================================================
DirectXBitmapBltWithTransparency16:   ;draws 16-bit bmp to screen
;============================================================
;USES TOP-LEFT PIXEL AS COLOR KEY
;params (1,2,...):
;dest. x
;dest. y
;width (of whole BMP)
;height
;pointer to bmp in memory
;use color key? set to NULL or DDBLT_KEYSRC
;image size
;flag (use 565 or 555)
;dest. RECT width
;dest. RECT height

push ebp
mov ebp,esp
push esi
push edi
push ebx

;NOTE: DirectX seems to completely refuse to create client memory surfaces, despite
;what the DirectX 6.1 documentation says. The example was tried unsuccessfully,
;along with every possible flags combinations. I debugged DDRAW.dll and found
;that when the lpSurface member was set, DirectDraw would always nullify the DDSD_PITCH
;flag, and then jump to a procedure which depended on this flag being set. Absolute
;nonsense. As a result, the bitmap data has to be 'manually' copied to a standard
;plain DDraw surface. Also, SetSurfaceDesc could not update the lpSurface member
;after creation because the method wasn't present in the DD surface interface (null?!?!?!).

	mov ebx,[ebp+8+(4*4)]

;setup DirectDraw surface for bitmap
	call GetDDSurfaceStructSize
	mov [bltclientsurface.dwSize],eax
	mov [bltclientsurface.dwFlags],DDSD_WIDTH+DDSD_HEIGHT+DDSD_PIXELFORMAT+\
				       DDSD_CKSRCBLT+DDSD_CAPS+DDSD_PITCH

	mov eax,[ebp+16]
	mov ecx,[ebp+20]
	mov [bltclientsurface.dwWidth],eax
	mov [bltclientsurface.dwHeight],ecx
	shl eax,1
	mov [bltclientsurface.lpSurface],ebx
	mov [bltclientsurface.lPitch],eax

	mov [bltclientsurface.ddpfPixelFormat.dwSize],sizeof.DDPIXELFORMAT
	mov [bltclientsurface.ddpfPixelFormat.dwFlags],DDPF_RGB
	mov [bltclientsurface.ddpfPixelFormat.dwRGBBitCount],16
	mov eax,[ebp+8+(7*4)]
	mov ecx,0000F800h
	mov edx,000007E0h
	cmp eax,SCREEN_16BITRGB_FORMAT_565
	     jz DirectXBitmapBltWithTransparency_Use565
	mov ecx,00007C00h
	mov edx,000003E0h
	DirectXBitmapBltWithTransparency_Use565:
	mov [bltclientsurface.ddpfPixelFormat.dwRBitMask],ecx
	mov [bltclientsurface.ddpfPixelFormat.dwGBitMask],edx
	mov [bltclientsurface.ddpfPixelFormat.dwBBitMask],0000001Fh

	mov [bltclientsurface.ddsCaps.dwCaps],DDSCAPS_SYSTEMMEMORY+DDSCAPS_OFFSCREENPLAIN

     ;get color key
	mov eax,[ebp+8+(6*4)]	     ;image size
	add eax,ebx		     ;end of image
	mov ecx,[bltclientsurface.dwWidth]
	add ecx,1
	and ecx,(not 1) 	;round up to nearest even width number (for scan line filling)
	shl ecx,2
	sub eax,ecx
	xor edx,edx
	mov dx,[eax]		;top-left pixel
	mov [bltclientsurface.ddckCKSrcBlt.dwColorSpaceLowValue],edx
	mov [bltclientsurface.ddckCKSrcBlt.dwColorSpaceHighValue],edx

;create client memory area DirectDraw surface
	call GetDDrawObject
	test eax,eax
	     jz DirectXBitmapBltWithTransparency_EXIT

	push 0
	push bltclientsurfacehandle
	push bltclientsurface
	mov ecx,[eax]
	push eax
	call DWORD [ecx+DirectDraw.CreateSurface]
	test eax,eax
	    jnz DirectXBitmapBltWithTransparency_EXIT

   ;lock surface
	call GetDDSurfaceStructSize
	mov [bltclientsurfaceretinfo.dwSize],eax
	mov ecx,[bltclientsurfacehandle]
	mov [bltclientsurfaceretinfo.dwFlags],DDSD_LPSURFACE+DDSD_PITCH+DDSD_WIDTH+DDSD_HEIGHT
	mov eax,[ecx]
	push 0
	push DDLOCK_READONLY+DDLOCK_SURFACEMEMORYPTR+DDLOCK_WAIT+DDLOCK_WRITEONLY
	push bltclientsurfaceretinfo
	push 0
	push ecx
	call DWORD [eax+DirectDrawSurface.Lock]
	test eax,eax
	     jnz DirectXBitmapBltWithTransparency_EXIT_RELEASE

   ;copy bitmap

	stdcall CopyBitmapBitsToSurface16,ebx,[bltclientsurfaceretinfo.lpSurface],\
		[bltclientsurfaceretinfo.dwWidth],[bltclientsurfaceretinfo.dwHeight],\
		[bltclientsurfaceretinfo.lPitch]

   ;unlock surface
	mov ecx,[bltclientsurfacehandle]
	push 0
	mov eax,[ecx]
	push ecx
	call DWORD [eax+DirectDrawSurface.Unlock]

   ;set RECTs
	xor eax,eax
	mov [bltclientsourcearea.left],eax
	mov [bltclientsourcearea.top],eax
	mov eax,[bltclientsurfaceretinfo.dwWidth]
	mov edx,[bltclientsurfaceretinfo.dwHeight]
	mov [bltclientsourcearea.right],eax
	mov [bltclientsourcearea.bottom],edx

	mov eax,[ebp+8]
	mov edx,[ebp+12]
	mov [bltclientdestarea.left],eax
	mov [bltclientdestarea.top],edx
	add eax,[ebp+4+(9*4)]		 ;dest. RECT width/height (for potential resize)
	add edx,[ebp+4+(10*4)]
	mov [bltclientdestarea.right],eax
	mov [bltclientdestarea.bottom],edx

    ;blit to screen
	mov [bltcientbltfx.dwSize],sizeof.DDBLTFX
	mov [bltcientbltfx.dwROP],SRCCOPY

	mov eax,[BaseDDSurfaceADDR]
	mov edx,[eax]
	mov ecx,[edx+4Ch]		;get DD surface for screen

	mov eax,[ebp+8+(5*4)]		;use color key?
	or eax,DDBLT_WAIT+DDBLT_ROP
	push bltcientbltfx
	push eax
	push bltclientsourcearea
	push [bltclientsurfacehandle]
	push bltclientdestarea
	mov eax,[ecx]
	push ecx
	call DWORD [eax+DirectDrawSurface.Blt]

DirectXBitmapBltWithTransparency_EXIT_RELEASE:

   ;release temp surface
	mov ecx,[bltclientsurfacehandle]
	mov eax,[ecx]
	push ecx
	call DWORD [eax+DirectDrawSurface.Release]

DirectXBitmapBltWithTransparency_EXIT:
pop ebx
pop edi
pop esi
mov esp,ebp
pop ebp
ret (10*4)

align 16
;============================================================
LoadRGBBitmapFromFile:	 ;draws to screen
;============================================================
;loads a bitmap from file. Returns a BITMAPINFO structure at
;specified address and bitmap data in a VirtualAlloc buffer.
;returns eax null if error or eax = buffer if success.

;if bitmap is not BI_RGB and min 8bpp or invalid file, is not loaded.

;param1 = name of file
;param2 = pointer for BITMAPINFOHEADER structure storage (may be larger to accomodate BITMAPINFO, not affected)
;param3 = if null, no effect; if a bitcount, only loads bitmap if it matches this bitcount

push ebp
mov ebp,esp
sub esp,20
push esi
push edi
push ebx

	xor ebx,ebx
	mov [ebp-4],ebx
	mov [ebp-16],ebx

    ;open potential bitmap file
	invoke CreateFile,[ebp+8],GENERIC_READ,FILE_SHARE_READ,0,\
			  OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
	cmp eax,-1
	mov ebx,eax
	    jz LoadRGBBitmapFromFile_ERROR

	invoke GetFileSize,ebx,0
	cmp eax,-1
	mov [ebp-12],eax
	    jz LoadRGBBitmapFromFile_ERROR
	cmp eax,(sizeof.BITMAPFILEHEADER+sizeof.BITMAPCOREHEADER)   ;~
	mov esi,eax
	    jb LoadRGBBitmapFromFile_ERROR

    ;cap header read size
	cmp esi,(sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER)
	    jb LoadRGBBitmapFromFile_SmallSizeSufficient
	    mov esi,(sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER)
	LoadRGBBitmapFromFile_SmallSizeSufficient:

    ;temp memory for header
	invoke LocalAlloc,LMEM_FIXED,esi
	test eax,eax
	mov [ebp-4],eax
	mov edi,eax
	    jz LoadRGBBitmapFromFile_ERROR

    ;read header
	invoke SetFilePointer,ebx,0,0,FILE_BEGIN
	cmp eax,-1
	    jz LoadRGBBitmapFromFile_ERROR
	lea edx,[ebp-8]
	invoke ReadFile,ebx,[ebp-4],esi,edx,0
	test eax,eax
	     jz LoadRGBBitmapFromFile_ERROR
	cmp esi,[ebp-8]
	    jnz LoadRGBBitmapFromFile_ERROR

    ;test file header
	mov eax,[edi+BITMAPFILEHEADER.bfSize]
	mov ecx,[edi+BITMAPFILEHEADER.bfOffBits]     ;ecx = offset of bitmap data in file
	cmp [edi+BITMAPFILEHEADER.bfType],'BM'
	    jnz LoadRGBBitmapFromFile_ERROR
	cmp eax,[ebp-12]
	    jnz LoadRGBBitmapFromFile_ERROR
	cmp ecx,eax
	    jnb LoadRGBBitmapFromFile_ERROR

    ;get info/core header
	mov edx,[ebp+12]
	add edi,sizeof.BITMAPFILEHEADER
	mov eax,[edi]
	cmp eax,sizeof.BITMAPINFOHEADER
	    jz LoadRGBBitmapFromFile_INFOHEADER
	cmp eax,sizeof.BITMAPCOREHEADER
	    jnz LoadRGBBitmapFromFile_ERROR

    ;core header
	cmp [edi+BITMAPCOREHEADER.bcBitCount],24     ;for core header, should be 24 bits for RGB
	    jnz LoadRGBBitmapFromFile_ERROR
	xor eax,eax
	xor ecx,ecx
	mov ax,[edi+BITMAPCOREHEADER.bcWidth]
	mov cx,[edi+BITMAPCOREHEADER.bcHeight]
	cmp eax,0
	mov [edx+BITMAPINFOHEADER.biWidth],eax
	    jle LoadRGBBitmapFromFile_ERROR
	cmp ecx,0
	mov [edx+BITMAPINFOHEADER.biHeight],ecx
	    jle LoadRGBBitmapFromFile_ERROR
	mov [edx+BITMAPINFOHEADER.biBitCount],24
	jmp LoadRGBBitmapFromFile_ENDINFOHEADER

    ;info header
	LoadRGBBitmapFromFile_INFOHEADER:

	cmp DWORD [ebp-12],(sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER) ;if info header, file size must correspond
	    jb LoadRGBBitmapFromFile_ERROR
	cmp [edi+BITMAPINFOHEADER.biCompression],BI_RGB
	    jnz LoadRGBBitmapFromFile_ERROR

	mov eax,[edi+BITMAPINFOHEADER.biWidth]
	mov ecx,[edi+BITMAPINFOHEADER.biHeight]
	cmp eax,0
	mov [edx+BITMAPINFOHEADER.biWidth],eax
	    jle LoadRGBBitmapFromFile_ERROR
	cmp ecx,0
	mov [edx+BITMAPINFOHEADER.biHeight],ecx
	    jle LoadRGBBitmapFromFile_ERROR
	mov ax,[edi+BITMAPINFOHEADER.biBitCount]
	cmp ax,8
	    jb LoadRGBBitmapFromFile_ERROR
	mov [edx+BITMAPINFOHEADER.biBitCount],ax
	LoadRGBBitmapFromFile_ENDINFOHEADER:

     ;if necessary, re-test bitcount (param3)
	mov eax,[ebp+16]
	test eax,eax
	     jz LoadRGBBitmapFromFile_NoForceBitCount
	cmp ax,[edx+BITMAPINFOHEADER.biBitCount]
	    jnz LoadRGBBitmapFromFile_ERROR
	LoadRGBBitmapFromFile_NoForceBitCount:

     ;finish dest. BITMAPINFOHEADER struct

       ;SizeImage is gotten from (filesize-offsetbits); cannot use depth*width*height
       ;because scan lines must be DWORD-address-aligned (when array of bits is
       ;loaded into memory at a DWORD-aligned location)
	mov eax,[ebp-12]   ;size
	sub eax,[edi-4]    ;offset to bits
	mov [edx+BITMAPINFOHEADER.biSizeImage],eax
	mov [ebp-20],eax
	mov [edx+BITMAPINFOHEADER.biSize],sizeof.BITMAPINFOHEADER
	xor eax,eax
	mov [edx+BITMAPINFOHEADER.biPlanes],1
	mov [edx+BITMAPINFOHEADER.biCompression],BI_RGB
	mov [edx+BITMAPINFOHEADER.biXPelsPerMeter],eax
	mov [edx+BITMAPINFOHEADER.biYPelsPerMeter],eax
	mov [edx+BITMAPINFOHEADER.biClrUsed],eax
	mov [edx+BITMAPINFOHEADER.biClrImportant],eax

    ;create virtual memory buffer
	invoke VirtualAlloc,0,[ebp-20],MEM_COMMIT+MEM_RESERVE,PAGE_READWRITE
	test eax,eax
	mov [ebp-16],eax
	    jz LoadRGBBitmapFromFile_ERROR

    ;summon bitmap data (from file offset for bitmap data)
	invoke SetFilePointer,ebx,[edi-4],0,FILE_BEGIN
	cmp eax,-1
	    jz LoadRGBBitmapFromFile_ERROR

	lea edx,[ebp-8]
	invoke ReadFile,ebx,[ebp-16],[ebp-20],edx,0
	test eax,eax
	mov ecx,[ebp-8]
	     jz LoadRGBBitmapFromFile_ERROR
	cmp ecx,[ebp-20]
	    jnz LoadRGBBitmapFromFile_ERROR

    ;free temp mem and release bmp file
	invoke LocalFree,[ebp-4]
	invoke CloseHandle,ebx
	mov eax,[ebp-16]

LoadRGBBitmapFromFile_EXIT:
pop ebx
pop edi
pop esi
mov esp,ebp
pop ebp
ret 12

	LoadRGBBitmapFromFile_ERROR:
	VIRTUALRELEASENONULL DWORD [ebp-16],DWORD [ebp-20]
	LOCALRELEASENONULL DWORD [ebp-4]
	CLOSEHANDLE ebx
	xor eax,eax
	jmp LoadRGBBitmapFromFile_EXIT

align 16
;============================================================
Convert24BitImageTo16Bit:
;============================================================
;NOT OPTIMIZED
;param1 = pointer to source bits (must be DWORD-aligned)
;param2 = pointer to dest bits (must be DWORD-aligned)
;param3 = 24-bit width
;param4 = 24-bit height
;param5 = SCREEN_16BITRGB_FORMAT_565 or SCREEN_16BITRGB_FORMAT_555
;returns new size

push ebp
mov ebp,esp
sub esp,4
push esi
push edi
push ebx

	mov esi,[ebp+4+(1*4)]
	mov edi,[ebp+4+(2*4)]

	mov DWORD [ebp-4],Convert24BitPixelTo16Bit_565
	cmp DWORD [ebp+4+(5*4)],SCREEN_16BITRGB_FORMAT_565
	    jz Convert24BitImageTo16Bit_use565
	    mov DWORD [ebp-4],Convert24BitPixelTo16Bit_555
	Convert24BitImageTo16Bit_use565:

	Convert24BitImageTo16Bit_LOOP_HEIGHT:
	mov ebx,[ebp+4+(3*4)]	   ;width

		Convert24BitImageTo16Bit_LOOP_WIDTH:
		call DWORD [ebp-4]	;convert pixel at esi
		mov [edi],ax
		add esi,3
		add edi,2
		dec ebx
		    jnz Convert24BitImageTo16Bit_LOOP_WIDTH

       ;align pointers at DWORD boundaries for next scanline
	add esi,3		       ;if address has any of the two lowest bits set, will round up
	add edi,3
	and esi,(not 3)
	and edi,(not 3)
	dec DWORD [ebp+4+(4*4)]        ;height
	    jnz Convert24BitImageTo16Bit_LOOP_HEIGHT

	mov eax,edi
	sub eax,[ebp+4+(2*4)]	       ;16-bit image size

pop ebx
pop edi
pop esi
mov esp,ebp
pop ebp
ret 20

align 16
;============================================================
Convert24BitPixelTo16Bit_565:	 ;esi = pointer to pixel, ax = 16-bit
;============================================================

	mov ax,[esi]
	mov ecx,eax
	mov dh,[esi+2]
	and eax,0F8h
	and ecx,0FC00h
	and edx,0F800h
	shr eax,3
	shr ecx,5
	or eax,ecx
	or eax,edx

ret

align 16
;============================================================
Convert24BitPixelTo16Bit_555:	 ;esi = pointer to pixel, ax = 16-bit
;============================================================
	mov ax,[esi]
	mov ecx,eax
	mov dh,[esi+2]
	and eax,0F8h
	and ecx,0F800h
	and edx,0F800h
	shr eax,3
	shr ecx,6
	shr edx,1
	or eax,ecx
	or eax,edx
ret

align 16
;============================================================
CopyBitmapBitsToSurface16:
;============================================================
;also flips image vertically for screen draw
;1 - bits pointer
;2 - surface pointer
;3 - width
;4 - height
;5 - pitch

push ebp
mov ebp,esp
sub esp,4
push esi
push edi
push ebx


	mov esi,[ebp+4+(1*4)]
	mov edi,[ebp+4+(2*4)]

	mov eax,[ebp+4+(5*4)]	 ;pitch
	mov ecx,[ebp+4+(4*4)]	 ;height
	mul ecx
	add edi,eax
	mov [ebp-4],edi 	 ;[ebp-4] points to end of surface in memory

	CopyBitmapBitsToSurface16_LOOP_HEIGHT:
	mov ebx,[ebp+4+(3*4)]
	mov edi,[ebp-4]
	sub edi,[ebp+4+(5*4)]
	mov [ebp-4],edi

		CopyBitmapBitsToSurface16_LOOP_WIDTH:
		mov ecx,ebx
		and ecx,(not 1)
		test ecx,ecx
		     jz CopyBitmapBitsToSurface16_LOOP_WIDTH_ONEPIXEL
		shr ecx,1
		rep movsd
		CopyBitmapBitsToSurface16_LOOP_WIDTH_ONEPIXEL:

		test ebx,1
		     jz CopyBitmapBitsToSurface16_LOOP_WIDTH_EVEN
		movsw
		CopyBitmapBitsToSurface16_LOOP_WIDTH_EVEN:

	add esi,3	   ;align source bits to DWORD
	and esi,(not 3)

	dec DWORD [ebp+4+(4*4)]
	    jnz CopyBitmapBitsToSurface16_LOOP_HEIGHT

pop ebx
pop edi
pop esi
mov esp,ebp
pop ebp
ret 20

align 16
;============================================================
Load8BitPaletteBitmapFromFile:
;============================================================

;param1 = name of file
;param2 = pointer for BITMAPINFOHEADER structure storage (may be larger to accomodate BITMAPINFO, not affected)
;param3 = pointer for palette data
;param4 = pointer for bitmap data or NULL to return
;param5 = number of bytes between each scanline. if 0, use bitmap width
;must be 8bpp and dimensions must be power of 2 and must be at least 4x4d, uncompressed
;returns eax = bitmap data or null if failure

push ebp
mov ebp,esp
sub esp,16
push esi
push edi
push ebx

	xor ebx,ebx
	mov [ebp-16],ebx

    ;open potential bitmap file
	invoke CreateFile,[param1],GENERIC_READ,FILE_SHARE_READ,0,\
			  OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
	cmp eax,-1
	mov ebx,eax
	    jz Load8BitPaletteBitmapFromFile_ERROR

	invoke GetFileSize,ebx,0
	cmp eax,-1
	mov [ebp-12],eax
	    jz Load8BitPaletteBitmapFromFile_ERROR
	cmp eax,(sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER+4*256+4*4)
	mov esi,eax
	    jb Load8BitPaletteBitmapFromFile_ERROR

    ;read file header only
	mov edi,stack_ext_buffer
	READFROMFILE ebx,0,sizeof.BITMAPFILEHEADER,edi,Load8BitPaletteBitmapFromFile_ERROR

    ;test file header
	mov eax,[edi+BITMAPFILEHEADER.bfSize]
	mov ecx,[edi+BITMAPFILEHEADER.bfOffBits]     ;ecx = offset of bitmap data in file
	cmp [edi+BITMAPFILEHEADER.bfType],'BM'
	    jnz Load8BitPaletteBitmapFromFile_ERROR
	cmp eax,[ebp-12]
	    jnz Load8BitPaletteBitmapFromFile_ERROR
	cmp ecx,eax
	    jnb Load8BitPaletteBitmapFromFile_ERROR
	mov [ebp-8],ecx

    ;get info/core header
	mov edi,[param2]
	READFROMFILE ebx,sizeof.BITMAPFILEHEADER,sizeof.BITMAPINFOHEADER,edi,Load8BitPaletteBitmapFromFile_ERROR

    ;check info header

	mov eax,[edi+BITMAPINFOHEADER.biWidth]
	cmp eax,4
	    jl Load8BitPaletteBitmapFromFile_ERROR
	cdq
	mov ecx,2
	div ecx
	test edx,edx
	     jnz Load8BitPaletteBitmapFromFile_ERROR

	mov eax,[edi+BITMAPINFOHEADER.biHeight]
	cmp eax,4
	    jl Load8BitPaletteBitmapFromFile_ERROR
	cdq
	mov ecx,2
	div ecx
	test edx,edx
	     jnz Load8BitPaletteBitmapFromFile_ERROR

	cmp [edi+BITMAPINFOHEADER.biBitCount],8
	    jnz Load8BitPaletteBitmapFromFile_ERROR

	mov eax,[edi+BITMAPINFOHEADER.biWidth]
	mov ecx,[edi+BITMAPINFOHEADER.biHeight]
	mul ecx
	mov [edi+BITMAPINFOHEADER.biSizeImage],eax  ;set image size if not set
	add ecx,(sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER+4*256)
	cmp [ebp-12],ecx
	    jb Load8BitPaletteBitmapFromFile_ERROR

    ;read palette into buffer

	READFROMFILE ebx,sizeof.BITMAPFILEHEADER+sizeof.BITMAPINFOHEADER,\
		     (4*256),[param3],Load8BitPaletteBitmapFromFile_ERROR

    ;create virtual memory buffer if necessary
	mov eax,[param4]
	test eax,eax
	     jnz Load8BitPaletteBitmapFromFile_NoMemAlloc
	mov esi,[edi+BITMAPINFOHEADER.biSizeImage]
	mov ecx,[param5]
	test ecx,ecx
	     jz Load8BitPaletteBitmapFromFile_NoScanlineSizeAdjust
	mov eax,[edi+BITMAPINFOHEADER.biHeight]
	mul ecx
	mov esi,eax
	Load8BitPaletteBitmapFromFile_NoScanlineSizeAdjust:
	invoke VirtualAlloc,0,esi,MEM_COMMIT+MEM_RESERVE,PAGE_READWRITE
	test eax,eax
	    jz Load8BitPaletteBitmapFromFile_ERROR
	Load8BitPaletteBitmapFromFile_NoMemAlloc:
	mov [ebp-16],eax

    ;summon bitmap data (from file offset for bitmap data)

	cmp [param5],0
	    jz Load8BitPaletteBitmapFromFile_NoScanlineOffset

	invoke SetFilePointer,ebx,[ebp-8],0,FILE_BEGIN
	cmp eax,-1
	    jz Load8BitPaletteBitmapFromFile_ERROR

	mov eax,[edi+BITMAPINFOHEADER.biHeight]
	mov esi,[ebp-16]
	mov [ebp-8],eax
	Load8BitPaletteBitmapFromFile_DoScanlineOffsetLoop:
	sub esp,4
	mov ecx,esp
	invoke ReadFile,ebx,esi,[edi+BITMAPINFOHEADER.biWidth],ecx,0
	pop ecx
	test eax,eax
	     jz Load8BitPaletteBitmapFromFile_ERROR
	cmp ecx,[edi+BITMAPINFOHEADER.biWidth]
	    jnz Load8BitPaletteBitmapFromFile_ERROR
	add esi,[param5]
	dec DWORD [ebp-8]
	    jnz Load8BitPaletteBitmapFromFile_DoScanlineOffsetLoop

	jmp Load8BitPaletteBitmapFromFile_ExitScanLineOffset
	Load8BitPaletteBitmapFromFile_NoScanlineOffset:
	READFROMFILE ebx,[ebp-8],[edi+BITMAPINFOHEADER.biSizeImage],[ebp-16],Load8BitPaletteBitmapFromFile_ERROR
	Load8BitPaletteBitmapFromFile_ExitScanLineOffset:

    ;free temp mem and release bmp file
	invoke CloseHandle,ebx
	mov eax,[ebp-16]

Load8BitPaletteBitmapFromFile_EXIT:
pop ebx
pop edi
pop esi
mov esp,ebp
pop ebp
ret 20

Load8BitPaletteBitmapFromFile_ERROR:
	VIRTUALRELEASENONULL DWORD [ebp-16],[edi+BITMAPINFOHEADER.biSizeImage]
	CLOSEHANDLE ebx
	xor eax,eax
	jmp Load8BitPaletteBitmapFromFile_EXIT

align 16
;============================================================
QUEUE_Reset:	;ecx = STDQUEUEINFO, eax = max queue entries, edx = size of structure for each member
;============================================================
;set edx = eax = null if simply freeing mem
;eax = maxnodes = -1 if no practical limit

push ebx

	mov [ecx+STDQUEUEINFO.structsize],edx	;set new struct size already
	mov [ecx+STDQUEUEINFO.maxnodes],eax

	xor eax,eax
	mov ebx,[ecx+STDQUEUEINFO.newestnode]	;fetch first node
	mov [ecx+STDQUEUEINFO.nodecount],eax
	mov [ecx+STDQUEUEINFO.newestnode],eax
	mov [ecx+STDQUEUEINFO.oldestnode],eax

	test ebx,ebx
	     jz QUEUE_Reset_Exit

	QUEUE_Reset_Loop:
	mov ecx,ebx
	mov ebx,[ebx+STDQUEUENODE.oldernode]
	invoke LocalFree,ecx
	test ebx,ebx
	     jnz QUEUE_Reset_Loop

QUEUE_Reset_Exit:

pop ebx
ret

align 16
;============================================================
QUEUE_Add_Trunc:    ;ecx = STDQUEUEINFO, edx = struct
;============================================================
;if queue is full, removes last entry first
;returns eax = non-null if success

push esi
push edi

	mov esi,ecx
	mov edi,edx

;if queue full, first remove last node
	mov eax,[esi+STDQUEUEINFO.nodecount]
	cmp eax,[esi+STDQUEUEINFO.maxnodes]
	    jb QUEUE_Add_Trunc_AddAbsolute

	mov ecx,esi
	call QUEUE_DeleteLast

QUEUE_Add_Trunc_AddAbsolute:

	mov ecx,esi
	mov edx,edi
	call QUEUE_Add_Always	   ;same return value

pop edi
pop esi
ret



align 16
;============================================================
QUEUE_Add_Always:    ;ecx = STDQUEUEINFO, edx = struct
;============================================================
;ignores max node count
;returns eax = null if error
push esi
push edi
push ebx
	mov esi,ecx
	mov edi,edx

	mov eax,[esi+STDQUEUEINFO.structsize]
	add eax,sizeof.STDQUEUENODE
	invoke LocalAlloc,LMEM_FIXED,eax
	test eax,eax
	     jz QUEUE_Add_Always_Error

	mov ebx,eax

	cmp [esi+STDQUEUEINFO.nodecount],0
	    jz QUEUE_Add_Always_FirstNode

	inc [esi+STDQUEUEINFO.nodecount]

	mov edx,[esi+STDQUEUEINFO.newestnode]	 ;will be 2nd newest node now
	mov [edx+STDQUEUENODE.newernode],ebx
	mov [ebx+STDQUEUENODE.oldernode],edx
	mov [ebx+STDQUEUENODE.newernode],0
	mov [esi+STDQUEUEINFO.newestnode],ebx

	mov ecx,[esi+STDQUEUEINFO.structsize]
	mov esi,edi
	lea edi,[ebx+sizeof.STDQUEUENODE]
	shr ecx,2
	rep movsd

	mov eax,TRUE
pop ebx
pop edi
pop esi
ret

QUEUE_Add_Always_FirstNode:
	xor eax,eax
	mov [esi+STDQUEUEINFO.nodecount],1
	mov [esi+STDQUEUEINFO.newestnode],ebx
	mov [esi+STDQUEUEINFO.oldestnode],ebx

	mov [ebx+STDQUEUENODE.newernode],eax
	mov [ebx+STDQUEUENODE.oldernode],eax

	mov ecx,[esi+STDQUEUEINFO.structsize]
	mov esi,edi
	lea edi,[ebx+sizeof.STDQUEUENODE]
	shr ecx,2
	rep movsd

	mov eax,TRUE
pop ebx
pop edi
pop esi
ret


QUEUE_Add_Always_Error:
	xor eax,eax
pop ebx
pop edi
pop esi
ret

align 16
;============================================================
QUEUE_DeleteLast:    ;ecx = STDQUEUEINFO
;============================================================
;if empty, does nothing

push ebx
push esi

	mov ebx,[ecx+STDQUEUEINFO.oldestnode]
	mov esi,ecx

	test ebx,ebx			   ;if empty, do nothing
	     jz QUEUE_DeleteLast_Exit

	cmp [esi+STDQUEUEINFO.nodecount],1
	    jz QUEUE_DeleteLast_LastNode

	mov eax,[ebx+STDQUEUENODE.newernode]
	mov [esi+STDQUEUEINFO.oldestnode],eax		;2nd oldest node is now oldest
	mov DWORD [eax+STDQUEUENODE.oldernode],0	;no older node
	invoke LocalFree,ebx

	dec [esi+STDQUEUEINFO.nodecount]

QUEUE_DeleteLast_Exit:
pop esi
pop ebx
ret

QUEUE_DeleteLast_LastNode:
	xor eax,eax
	mov [esi+STDQUEUEINFO.newestnode],eax
	mov [esi+STDQUEUEINFO.oldestnode],eax
	mov [esi+STDQUEUEINFO.nodecount],eax
	invoke LocalFree,ebx
pop esi
pop ebx
ret

align 16
;============================================================
IsSoundInTPA:	 ;ecx = TPA struct, eax = sound identifier
;============================================================
;this is a wrap. returns eax = false if not found, true otherwise

push esi
add ecx,18h	;offset into TPA struct
mov esi,ecx

;local vars
push eax	 ;return var (also sound ID, don't know if necessary to be sound ID, but can't hurt)
push eax	 ;sound ID

;params (ecx already set)
lea edx,[esp+4]
push esp	 ;pointer to sound ID
push edx	 ;pointer to return var
call [TPACheckIDPresenceProc]		;returns pointer to return var in eax (same as param1)

mov eax,[eax]
cmp eax,[esi]
    jz IsSoundInTPA_NotFound
cmp DWORD [eax+14h],0
    jz IsSoundInTPA_NotFound

add esp,8	       ;local vars
mov eax,TRUE
pop esi
ret

IsSoundInTPA_NotFound:
add esp,8	       ;local vars
xor eax,eax
pop esi
ret


;============================================================
proc CheckFacingBetweenInstanceAndTarget  ;(on x-y plane only)
;============================================================
local FacingVector:vector,DistanceVector:vector,DotProductValue:DWORD

; esi = source location + quaternion
; edi = dest location (target)
; ebx = pointer to cosine of maximum allowed angle between source-dest vector and source instance orientation vector
; returns al = TRUE if within or equal to angle, FALSE otherwise


	;a = angle to get, b = limit angle, vectors are X-Y only (horizontal only)
	;                  -> assumed 0 <= b <= pi
	;-> limit angle is for each side of head, not total view angle
	;cos(a) = vector1.vector2 / (sizeof.vector1*sizeof.vector2)     ;. = dot product ; both vectors assumed non-zero
	;validity equation:      a <= b  (for any angle between 0 and 90)
	;                   cos(a) >= cos(b)
	;                   vector1.vector2 / (sizeof.vector1*sizeof.vector2) >= cos(b)
	;                   vector1.vector2 >= cos(b)*(sizeof.vector1*sizeof.vector2)      (valid since vector sizes will never change sign, always positive)
	;                   vector1.vector2 >= cos(b)*sizeof.vector1*sizeof.vector2
	;                   cos(b)*sizeof.vector1*sizeof.vector2 <= vector1.vector2

	lea ecx,[esi+sizeof.vector]
	lea edx,[FacingVector]
	stdcall CONVERT_TO_FACING_PROC_010,edx,0

	fld [edi+LOCATION.X]
	fsub [esi+LOCATION.X]
	fst [DistanceVector.X]
	fmul [FacingVector.X]
	fld [edi+LOCATION.Y]
	fsub [esi+LOCATION.Y]
	fst [DistanceVector.Y]
	fmul [FacingVector.Y]
	faddp st1,st0
	fstp [DotProductValue]

	fld [FacingVector.X]
	fmul st0,st0
	fld [FacingVector.Y]
	fmul st0,st0
	faddp st1,st0
	fsqrt

	fld [DistanceVector.X]
	fmul st0,st0
	fld [DistanceVector.Y]
	fmul st0,st0
	faddp st1,st0
	fsqrt

	fmulp st1,st0
	fld DWORD [ebx]
	fmulp st1,st0

	fcomp [DotProductValue]
	fstsw ax
	test ah,FPU_VALUE_BELOW_OR_EQUAL	 ;is cos(b)*sizeof.vector1*sizeof.vector2 <= vector1.vector2 ?
	mov eax,0
	setnz al				 ;if true, we're within angle

ret
endp

align 16
;============================================================
GetDistanceBetweenInstances:	;ecx = CInstance 1, edx = CInstance 2 (base addresses)
;============================================================
;uses CVector3
;RETURNS VALUE IN ST0

	fld [ecx+CInstance.Location.X]
	fsub [edx+CInstance.Location.X]
	fmul st0,st0
	fld [ecx+CInstance.Location.Y]
	fsub [edx+CInstance.Location.Y]
	fmul st0,st0
	fld [ecx+CInstance.Location.Z]
	fsub [edx+CInstance.Location.Z]
	fmul st0,st0
	faddp st1,st0
	faddp st1,st0
	fsqrt
ret


align 16
;============================================================
VectorMagnitude_st0:	      ;eax = pointer to vector, returns in st0
;============================================================

     fld DWORD [eax]
     fmul st0,st0
     fld DWORD [eax+4]
     fmul st0,st0
     fld DWORD [eax+8]
     fmul st0,st0
     faddp st1,st0
     faddp st1,st0
     fsqrt

ret






